/*
 * Copyright (c) 2014, 2014, Oracle and/or its affiliates. All rights reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */
package com.oracle.truffle.r.nodes.builtin.base;

import static com.oracle.truffle.r.nodes.builtin.RBuiltinKind.*;

import com.oracle.truffle.api.*;
import com.oracle.truffle.api.dsl.*;
import com.oracle.truffle.api.frame.*;
import com.oracle.truffle.api.frame.FrameInstance.FrameAccess;
import com.oracle.truffle.r.nodes.builtin.*;
import com.oracle.truffle.r.runtime.*;
import com.oracle.truffle.r.runtime.data.*;

/**
 * sys.R. See <a
 * href="https://stat.ethz.ch/R-manual/R-devel/library/base/html/sys.parent.html">here</a>. N.B. The
 * frame for the sys functions themselves is not counted in the R spec. Frames are numbered 0, 1, ..
 * starting from .GlobalEnv. Non-negative arguments are frame numbers, negative arguments are
 * relative to the current frame.
 */
public class FrameFunctions {

    public abstract static class FrameHelper extends RBuiltinNode {
        protected void error(String msg) throws RError {
            CompilerDirectives.transferToInterpreter();
            throw RError.getGenericError(getEncapsulatingSourceSection(), msg);
        }

        /**
         * Handles n > 0 and n < 0 and errors relating to stack depth.
         */
        protected Frame getFrame(int nn) {
            int n = nn;
            if (n > 0) {
                int d = Utils.stackDepth();
                if (n > d) {
                    error("not that many frames on the stack");
                }
                n = d - n;
            } else {
                n = -n + 1;
            }
            Frame callerFrame = Utils.getStackFrame(FrameAccess.MATERIALIZE, n);
            if (callerFrame == null) {
                error("not that many frames on the stack");
            }
            return callerFrame;
        }
    }

    @RBuiltin(name = "sys.nframe", kind = INTERNAL)
    public abstract static class SysNFrame extends RBuiltinNode {
        @Specialization()
        public int sysNFrame() {
            controlVisibility();
            return Utils.stackDepth();
        }
    }

    @RBuiltin(name = "sys.frame", kind = INTERNAL)
    public abstract static class SysFrame extends FrameHelper {
        @Specialization()
        public REnvironment sysFrame(int nd) {
            controlVisibility();
            int n = nd;
            if (n == 0) {
                // TODO Strictly this should be the value of .GlobalEnv
                // (which may differ from globalenv() during startup)
                return REnvironment.globalEnv();
            } else {
                Frame callerFrame = getFrame(n);
                return EnvFunctions.frameToEnvironment(callerFrame);
            }
        }

        @Specialization()
        public REnvironment sysFrame(double nd) {
            return sysFrame((int) nd);
        }
    }

    @RBuiltin(name = "sys.parent", kind = INTERNAL)
    public abstract static class SysParent extends RBuiltinNode {
        @Specialization()
        public int sysParent(int nd) {
            controlVisibility();
            int n = nd;
            int d = Utils.stackDepth();
            if (n > d) {
                return 0;
            } else {
                return d - n;
            }
        }

        @Specialization()
        public int sysParent(double nd) {
            return sysParent((int) nd);
        }
    }

    @RBuiltin(name = "sys.function", kind = INTERNAL)
    public abstract static class SysFunction extends FrameHelper {
        @Specialization()
        public Object sysFunction(int nd) {
            controlVisibility();
            int n = nd;
            // N.B. Despite the spec, n==0 is treated as the current function
            Frame callerFrame = getFrame(n);
            RFunction func = RArguments.getFunction(callerFrame);
            if (func == null) {
                return RNull.instance;
            } else {
                return func;
            }
        }

        @Specialization()
        public Object sysFunction(double nd) {
            return sysFunction((int) nd);
        }
    }

    @RBuiltin(name = "sys.parents", kind = INTERNAL)
    public abstract static class SysParents extends FrameHelper {
        @Specialization()
        public RIntVector sysParents() {
            controlVisibility();
            int d = Utils.stackDepth();
            int[] data = new int[d];
            for (int i = 0; i < d; i++) {
                data[i] = i;
            }
            return RDataFactory.createIntVector(data, RDataFactory.COMPLETE_VECTOR);
        }
    }

    @RBuiltin(name = "sys.frames", kind = INTERNAL)
    public abstract static class SysFrames extends FrameHelper {
        @Specialization()
        public Object sysFrames() {
            error("sys.frames is not implemented");
            return RNull.instance;
        }
    }

    /**
     * The environment of the caller of the function that called parent.frame.
     */
    @RBuiltin(name = "parent.frame", kind = INTERNAL)
    public abstract static class ParentFrame extends FrameHelper {
        @Specialization()
        public REnvironment parentFrame(double nd) {
            controlVisibility();
            int n = (int) nd;
            if (n == 0) {
                error("invalid 'n' value");
            }
            Frame callerFrame = Utils.getStackFrame(FrameAccess.MATERIALIZE, n + 1);
            if (callerFrame == null) {
                return REnvironment.globalEnv();
            } else {
                return EnvFunctions.frameToEnvironment(callerFrame);
            }
        }
    }
}
